from glotaran.builtin.megacomplexes.decay.decay_megacomplex import DecayMegacomplex
from glotaran.builtin.megacomplexes.decay.decay_parallel_megacomplex import (
    DecayParallelMegacomplex,
)
from glotaran.builtin.megacomplexes.decay.decay_sequential_megacomplex import (
    DecaySequentialMegacomplex,
)
