"""Package containing deprecated implementations which were removed.

To keep things organized the filenames should be like the
relative import path from glotaran root, but with ``_`` instead of ``.``.
E.g. ``glotaran.analysis.scheme`` would map to ``analysis_scheme.py``

The only exceptions to this rule are the root ``__init__.py`` which
is named ``glotaran_root.py`` and testing changed imports which should
be placed in ``test_changed_imports.py``.
"""
